# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import pytest

from airflow.providers.fab.auth_manager.schemas.role_and_permission_schema import (
    RoleCollection,
    role_collection_schema,
    role_schema,
)
from airflow.providers.fab.www.security import permissions

from unit.fab.auth_manager.api_endpoints.api_connexion_utils import create_role, delete_role

pytestmark = pytest.mark.db_test


class TestRoleCollectionItemSchema:
    @pytest.fixture(scope="class")
    def role(self, minimal_app_for_auth_api):
        with minimal_app_for_auth_api.app_context():
            yield create_role(
                minimal_app_for_auth_api,
                name="Test",
                permissions=[
                    (permissions.ACTION_CAN_CREATE, permissions.RESOURCE_CONNECTION),
                ],
            )
            delete_role(minimal_app_for_auth_api, "Test")

    @pytest.fixture(autouse=True)
    def _set_attrs(self, minimal_app_for_auth_api, role):
        self.app = minimal_app_for_auth_api
        self.role = role

    def test_serialize(self):
        deserialized_role = role_schema.dump(self.role)
        assert deserialized_role == {
            "name": "Test",
            "actions": [{"resource": {"name": "Connections"}, "action": {"name": "can_create"}}],
        }

    def test_deserialize(self):
        role = {
            "name": "Test",
            "actions": [{"resource": {"name": "Connections"}, "action": {"name": "can_create"}}],
        }
        role_obj = role_schema.load(role)
        assert role_obj == {
            "name": "Test",
            "permissions": [{"resource": {"name": "Connections"}, "action": {"name": "can_create"}}],
        }


class TestRoleCollectionSchema:
    def test_serialize(self, minimal_app_for_auth_api):
        with minimal_app_for_auth_api.app_context():
            role1 = create_role(
                minimal_app_for_auth_api,
                name="Test1",
                permissions=[
                    (permissions.ACTION_CAN_CREATE, permissions.RESOURCE_CONNECTION),
                ],
            )
            role2 = create_role(
                minimal_app_for_auth_api,
                name="Test2",
                permissions=[
                    (permissions.ACTION_CAN_EDIT, permissions.RESOURCE_DAG),
                ],
            )

            instance = RoleCollection([role1, role2], total_entries=2)
            deserialized = role_collection_schema.dump(instance)
            assert deserialized == {
                "roles": [
                    {
                        "name": "Test1",
                        "actions": [{"resource": {"name": "Connections"}, "action": {"name": "can_create"}}],
                    },
                    {
                        "name": "Test2",
                        "actions": [{"resource": {"name": "DAGs"}, "action": {"name": "can_edit"}}],
                    },
                ],
                "total_entries": 2,
            }

            delete_role(minimal_app_for_auth_api, "Test1")
            delete_role(minimal_app_for_auth_api, "Test2")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import msgspec
from flask import request
from flask.sessions import SecureCookieSessionInterface
from flask_babel import LazyString
from flask_session.sqlalchemy import SqlAlchemySessionInterface


class _LazySafeSerializer:
    def dumps(self, session_dict):
        encoder = msgspec.msgpack.Encoder(
            enc_hook=lambda obj: str(obj) if isinstance(obj, LazyString) else obj
        )
        return encoder.encode(dict(session_dict))

    def loads(self, data):
        decoder = msgspec.msgpack.Decoder()
        return decoder.decode(data)

    # optional old API
    encode = dumps
    decode = loads


class SessionExemptMixin:
    """Exempt certain blueprints/paths from autogenerated sessions."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.serializer = _LazySafeSerializer()

    def save_session(self, *args, **kwargs):
        """Prevent creating session from REST API and health requests."""
        if request.path == "/health":
            return None
        return super().save_session(*args, **kwargs)


class AirflowDatabaseSessionInterface(SessionExemptMixin, SqlAlchemySessionInterface):
    """Session interface that exempts some routes and stores session data in the database."""


class AirflowSecureCookieSessionInterface(SessionExemptMixin, SecureCookieSessionInterface):
    """Session interface that exempts some routes and stores session data in a signed cookie."""



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from airflow.providers.fab.auth_manager.cli_commands.definition import (
    ROLES_COMMANDS,
    SYNC_PERM_COMMAND,
    USERS_COMMANDS,
)


class TestCliDefinition:
    def test_users_commands(self):
        assert len(USERS_COMMANDS) == 8

    def test_roles_commands(self):
        assert len(ROLES_COMMANDS) == 7

    def test_sync_perm_command(self):
        assert SYNC_PERM_COMMAND.name == "sync-perm"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import types
from unittest.mock import MagicMock, patch

import pytest
from fastapi import HTTPException
from sqlalchemy import column

from airflow.providers.fab.auth_manager.api_fastapi.services.roles import FABAuthManagerRoles


@pytest.fixture
def fab_auth_manager():
    return MagicMock()


@pytest.fixture
def security_manager():
    sm = MagicMock()
    sm.get_action.side_effect = lambda n: object() if n in {"can_read", "can_edit"} else None
    sm.get_resource.side_effect = lambda n: object() if n in {"DAG"} else None
    return sm


def _make_role_obj(name: str, perms: list[tuple[str, str]]):
    perm_objs = [
        types.SimpleNamespace(
            action=types.SimpleNamespace(name=a),
            resource=types.SimpleNamespace(name=r),
        )
        for (a, r) in perms
    ]
    return types.SimpleNamespace(id=1, name=name, permissions=perm_objs)


class _FakeScalarCount:
    def __init__(self, value: int):
        self._value = value

    def one(self) -> int:
        return self._value


class _FakeScalarRoles:
    def __init__(self, items):
        self._items = items
        self._unique_called = False

    def unique(self):
        self._unique_called = True
        return self

    def all(self):
        return self._items


@patch("airflow.providers.fab.auth_manager.api_fastapi.services.roles.get_fab_auth_manager")
class TestRolesService:
    def setup_method(self):
        self.body_ok = types.SimpleNamespace(
            name="roleA",
            permissions=[
                types.SimpleNamespace(
                    action=types.SimpleNamespace(name="can_read"),
                    resource=types.SimpleNamespace(name="DAG"),
                )
            ],
        )

        self.body_bad_action = types.SimpleNamespace(
            name="roleB",
            permissions=[
                types.SimpleNamespace(
                    action=types.SimpleNamespace(name="no_such_action"),
                    resource=types.SimpleNamespace(name="DAG"),
                )
            ],
        )
        self.body_bad_resource = types.SimpleNamespace(
            name="roleC",
            permissions=[
                types.SimpleNamespace(
                    action=types.SimpleNamespace(name="can_read"),
                    resource=types.SimpleNamespace(name="NOPE"),
                )
            ],
        )

    # POST /roles

    def test_create_role_success(self, get_fab_auth_manager, fab_auth_manager, security_manager):
        security_manager.find_role.side_effect = [
            None,
            _make_role_obj("roleA", [("can_read", "DAG")]),
        ]
        fab_auth_manager.security_manager = security_manager
        get_fab_auth_manager.return_value = fab_auth_manager

        out = FABAuthManagerRoles.create_role(self.body_ok)

        assert out.name == "roleA"
        assert out.permissions
        assert out.permissions[0].action.name == "can_read"
        assert out.permissions[0].resource.name == "DAG"
        security_manager.bulk_sync_roles.assert_called_once_with(
            [{"role": "roleA", "perms": [("can_read", "DAG")]}]
        )

    def test_create_role_conflict(self, get_fab_auth_manager, fab_auth_manager, security_manager):
        security_manager.find_role.return_value = object()
        fab_auth_manager.security_manager = security_manager
        get_fab_auth_manager.return_value = fab_auth_manager

        with pytest.raises(HTTPException) as ex:
            FABAuthManagerRoles.create_role(self.body_ok)
        assert ex.value.status_code == 409

    def test_create_role_action_not_found(self, get_fab_auth_manager, fab_auth_manager, security_manager):
        security_manager.find_role.return_value = None
        fab_auth_manager.security_manager = security_manager
        get_fab_auth_manager.return_value = fab_auth_manager

        with pytest.raises(HTTPException) as ex:
            FABAuthManagerRoles.create_role(self.body_bad_action)
        assert ex.value.status_code == 400
        assert "action" in ex.value.detail

    def test_create_role_resource_not_found(self, get_fab_auth_manager, fab_auth_manager, security_manager):
        security_manager.find_role.return_value = None
        fab_auth_manager.security_manager = security_manager
        get_fab_auth_manager.return_value = fab_auth_manager

        with pytest.raises(HTTPException) as ex:
            FABAuthManagerRoles.create_role(self.body_bad_resource)
        assert ex.value.status_code == 400
        assert "resource" in ex.value.detail

    def test_create_role_unexpected_no_created(
        self, get_fab_auth_manager, fab_auth_manager, security_manager
    ):
        security_manager.find_role.side_effect = [None, None]
        fab_auth_manager.security_manager = security_manager
        get_fab_auth_manager.return_value = fab_auth_manager

        with pytest.raises(HTTPException) as ex:
            FABAuthManagerRoles.create_role(self.body_ok)
        assert ex.value.status_code == 500

    # GET /roles

    @patch("airflow.providers.fab.auth_manager.api_fastapi.services.roles.build_ordering")
    def test_get_roles_happy_path(self, build_ordering, get_fab_auth_manager):
        role1 = _make_role_obj("viewer", [("can_read", "DAG")])
        role2 = _make_role_obj("admin", [("can_read", "DAG")])
        fake_roles_result = _FakeScalarRoles([role1, role2])

        session = MagicMock()
        session.scalars.side_effect = [
            _FakeScalarCount(2),
            fake_roles_result,
        ]

        fab_auth_manager = MagicMock()
        fab_auth_manager.security_manager = MagicMock(session=session)
        get_fab_auth_manager.return_value = fab_auth_manager

        build_ordering.return_value = column("name").desc()

        out = FABAuthManagerRoles.get_roles(order_by="-name", limit=5, offset=3)

        assert out.total_entries == 2
        assert [r.name for r in out.roles] == ["viewer", "admin"]
        assert fake_roles_result._unique_called is True

        build_ordering.assert_called_once()
        args, kwargs = build_ordering.call_args
        assert args[0] == "-name"
        assert set(kwargs["allowed"].keys()) == {"name", "role_id"}

        assert session.scalars.call_count == 2

    @patch("airflow.providers.fab.auth_manager.api_fastapi.services.roles.build_ordering")
    def test_get_roles_invalid_order_by_bubbles_400(self, build_ordering, get_fab_auth_manager):
        session = MagicMock()
        fab_auth_manager = MagicMock()
        fab_auth_manager.security_manager = MagicMock(session=session)
        get_fab_auth_manager.return_value = fab_auth_manager

        build_ordering.side_effect = HTTPException(status_code=400, detail="disallowed")

        with pytest.raises(HTTPException) as ex:
            FABAuthManagerRoles.get_roles(order_by="nope", limit=10, offset=0)
        assert ex.value.status_code == 400

    # DELETE /roles/{name}

    def test_delete_role_success(self, get_fab_auth_manager, fab_auth_manager, security_manager):
        security_manager.find_role.return_value = _make_role_obj("roleA", [])
        fab_auth_manager.security_manager = security_manager
        get_fab_auth_manager.return_value = fab_auth_manager

        FABAuthManagerRoles.delete_role(name="roleA")

        security_manager.delete_role.assert_called_once()

    def test_delete_role_not_found(self, get_fab_auth_manager, fab_auth_manager, security_manager):
        security_manager.find_role.return_value = None
        fab_auth_manager.security_manager = security_manager
        get_fab_auth_manager.return_value = fab_auth_manager

        with pytest.raises(HTTPException) as ex:
            FABAuthManagerRoles.delete_role(name="roleA")
        assert ex.value.status_code == 404

    # GET /roles/{name}

    def test_get_role_success(self, get_fab_auth_manager, fab_auth_manager, security_manager):
        security_manager.find_role.return_value = _make_role_obj("roleA", [("can_read", "DAG")])
        fab_auth_manager.security_manager = security_manager
        get_fab_auth_manager.return_value = fab_auth_manager

        out = FABAuthManagerRoles.get_role(name="roleA")

        assert out.name == "roleA"
        assert out.permissions
        assert out.permissions[0].action.name == "can_read"
        assert out.permissions[0].resource.name == "DAG"

    def test_get_role_not_found(self, get_fab_auth_manager, fab_auth_manager, security_manager):
        security_manager.find_role.return_value = None
        fab_auth_manager.security_manager = security_manager
        get_fab_auth_manager.return_value = fab_auth_manager

        with pytest.raises(HTTPException) as ex:
            FABAuthManagerRoles.get_role(name="roleA")
        assert ex.value.status_code == 404

    # PATCH /roles/{name}

    def test_patch_role_success(self, get_fab_auth_manager, fab_auth_manager, security_manager):
        role = _make_role_obj("viewer", [("can_read", "DAG")])
        security_manager.find_role.return_value = role
        fab_auth_manager.security_manager = security_manager
        get_fab_auth_manager.return_value = fab_auth_manager
        body = types.SimpleNamespace(
            name="viewer",
            permissions=[
                types.SimpleNamespace(
                    action=types.SimpleNamespace(name="can_edit"),
                    resource=types.SimpleNamespace(name="DAG"),
                )
            ],
        )
        out = FABAuthManagerRoles.patch_role(body=body, name="viewer")
        assert out.name == "viewer"
        assert out.permissions
        assert out.permissions[0].action.name == "can_edit"
        assert out.permissions[0].resource.name == "DAG"

    def test_patch_role_rename_success(self, get_fab_auth_manager, fab_auth_manager, security_manager):
        role = _make_role_obj("viewer", [("can_edit", "DAG")])
        security_manager.find_role.return_value = role
        fab_auth_manager.security_manager = security_manager
        get_fab_auth_manager.return_value = fab_auth_manager
        body = types.SimpleNamespace(
            name="editor",
            permissions=[
                types.SimpleNamespace(
                    action=types.SimpleNamespace(name="can_edit"),
                    resource=types.SimpleNamespace(name="DAG"),
                )
            ],
        )
        out = FABAuthManagerRoles.patch_role(body=body, name="viewer")
        assert out.name == "editor"
        assert out.permissions
        assert out.permissions[0].action.name == "can_edit"
        assert out.permissions[0].resource.name == "DAG"

    def test_patch_role_with_update_mask(self, get_fab_auth_manager, fab_auth_manager, security_manager):
        role = _make_role_obj("viewer", [("can_read", "DAG")])
        security_manager.find_role.return_value = role
        fab_auth_manager.security_manager = security_manager
        get_fab_auth_manager.return_value = fab_auth_manager
        body = types.SimpleNamespace(
            name="viewer1",
            permissions=[
                types.SimpleNamespace(
                    action=types.SimpleNamespace(name="can_edit"),
                    resource=types.SimpleNamespace(name="DAG"),
                )
            ],
        )
        out = FABAuthManagerRoles.patch_role(
            body=body,
            name="viewer",
            update_mask=["actions"],
        )
        assert out.name == "viewer"
        assert out.permissions
        assert out.permissions[0].action.name == "can_edit"
        assert out.permissions[0].resource.name == "DAG"

    def test_patch_role_rename_with_update_mask(
        self, get_fab_auth_manager, fab_auth_manager, security_manager
    ):
        role = _make_role_obj("viewer", [("can_read", "DAG")])
        security_manager.find_role.return_value = role
        fab_auth_manager.security_manager = security_manager
        get_fab_auth_manager.return_value = fab_auth_manager
        body = types.SimpleNamespace(
            name="viewer1",
            permissions=[
                types.SimpleNamespace(
                    action=types.SimpleNamespace(name="can_edit"),
                    resource=types.SimpleNamespace(name="DAG"),
                )
            ],
        )
        out = FABAuthManagerRoles.patch_role(
            body=body,
            name="viewer",
            update_mask=["name"],
        )
        assert out.name == "viewer1"
        assert out.permissions
        assert out.permissions[0].action.name == "can_read"
        assert out.permissions[0].resource.name == "DAG"

    def test_patch_role_not_found(self, get_fab_auth_manager, fab_auth_manager, security_manager):
        security_manager.find_role.return_value = None
        fab_auth_manager.security_manager = security_manager
        get_fab_auth_manager.return_value = fab_auth_manager
        body = types.SimpleNamespace(
            name="viewer",
            permissions=[
                types.SimpleNamespace(
                    action=types.SimpleNamespace(name="can_edit"),
                    resource=types.SimpleNamespace(name="DAG"),
                )
            ],
        )
        with pytest.raises(HTTPException) as ex:
            FABAuthManagerRoles.patch_role(body=body, name="viewer")
        assert ex.value.status_code == 404



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import logging
from typing import TYPE_CHECKING
from unittest.mock import patch

if TYPE_CHECKING:
    import pytest

from airflow.providers.fab.auth_manager.api_fastapi.parameters import get_effective_limit


class TestParametersDependency:
    @patch("airflow.providers.fab.auth_manager.api_fastapi.parameters.conf")
    def test_get_effective_limit_uses_fallback_when_zero(self, conf_mock):
        conf_mock.getint.side_effect = lambda section, option: {
            "maximum_page_limit": 100,
            "fallback_page_limit": 25,
        }[option]

        limit_dep = get_effective_limit()

        assert limit_dep(limit=0) == 25
        conf_mock.getint.assert_any_call("api", "maximum_page_limit")
        conf_mock.getint.assert_any_call("api", "fallback_page_limit")

    @patch("airflow.providers.fab.auth_manager.api_fastapi.parameters.conf")
    def test_get_effective_limit_clamps_to_max_and_logs_warning(
        self, conf_mock, caplog: pytest.LogCaptureFixture
    ):
        conf_mock.getint.side_effect = lambda section, option: {
            "maximum_page_limit": 50,
            "fallback_page_limit": 20,
        }[option]

        limit_dep = get_effective_limit()

        caplog.set_level(logging.WARNING, logger="airflow.providers.fab.auth_manager.api_fastapi.parameters")

        result = limit_dep(limit=1000)

        assert result == 50
        assert any("exceeds the configured maximum page limit" in rec.getMessage() for rec in caplog.records)

    @patch("airflow.providers.fab.auth_manager.api_fastapi.parameters.conf")
    def test_get_effective_limit_returns_given_value_when_within_range(self, conf_mock):
        conf_mock.getint.side_effect = lambda section, option: {
            "maximum_page_limit": 200,
            "fallback_page_limit": 10,
        }[option]

        limit_dep = get_effective_limit()

        assert limit_dep(limit=150) == 150



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING

from fastapi import Depends, Path, Query, status

from airflow.api_fastapi.common.router import AirflowRouter
from airflow.api_fastapi.core_api.openapi.exceptions import create_openapi_http_exception_doc
from airflow.providers.fab.auth_manager.api_fastapi.datamodels.roles import (
    RoleBody,
    RoleCollectionResponse,
    RoleResponse,
)
from airflow.providers.fab.auth_manager.api_fastapi.parameters import get_effective_limit
from airflow.providers.fab.auth_manager.api_fastapi.security import requires_fab_custom_view
from airflow.providers.fab.auth_manager.api_fastapi.services.roles import FABAuthManagerRoles
from airflow.providers.fab.auth_manager.cli_commands.utils import get_application_builder
from airflow.providers.fab.www.security import permissions

if TYPE_CHECKING:
    from airflow.providers.fab.auth_manager.api_fastapi.datamodels.roles import (
        RoleBody,
        RoleResponse,
    )


roles_router = AirflowRouter(prefix="/fab/v1", tags=["FabAuthManager"])


@roles_router.post(
    "/roles",
    responses=create_openapi_http_exception_doc(
        [
            status.HTTP_400_BAD_REQUEST,
            status.HTTP_401_UNAUTHORIZED,
            status.HTTP_403_FORBIDDEN,
            status.HTTP_409_CONFLICT,
            status.HTTP_500_INTERNAL_SERVER_ERROR,
        ]
    ),
    dependencies=[Depends(requires_fab_custom_view("POST", permissions.RESOURCE_ROLE))],
)
def create_role(body: RoleBody) -> RoleResponse:
    """Create a new role (actions can be empty)."""
    with get_application_builder():
        return FABAuthManagerRoles.create_role(body=body)


@roles_router.get(
    "/roles",
    response_model=RoleCollectionResponse,
    responses=create_openapi_http_exception_doc(
        [
            status.HTTP_400_BAD_REQUEST,
            status.HTTP_401_UNAUTHORIZED,
            status.HTTP_403_FORBIDDEN,
            status.HTTP_500_INTERNAL_SERVER_ERROR,
        ]
    ),
    dependencies=[Depends(requires_fab_custom_view("GET", permissions.RESOURCE_ROLE))],
)
def get_roles(
    order_by: str = Query("name", description="Field to order by. Prefix with '-' for descending."),
    limit: int = Depends(get_effective_limit()),
    offset: int = Query(0, ge=0, description="Number of items to skip before starting to collect results."),
) -> RoleCollectionResponse:
    """List roles with pagination and ordering."""
    with get_application_builder():
        return FABAuthManagerRoles.get_roles(order_by=order_by, limit=limit, offset=offset)


@roles_router.delete(
    "/roles/{name}",
    responses=create_openapi_http_exception_doc(
        [
            status.HTTP_401_UNAUTHORIZED,
            status.HTTP_403_FORBIDDEN,
            status.HTTP_404_NOT_FOUND,
        ]
    ),
    dependencies=[Depends(requires_fab_custom_view("DELETE", permissions.RESOURCE_ROLE))],
)
def delete_role(name: str = Path(..., min_length=1)) -> None:
    """Delete an existing role."""
    with get_application_builder():
        return FABAuthManagerRoles.delete_role(name=name)


@roles_router.get(
    "/roles/{name}",
    responses=create_openapi_http_exception_doc(
        [status.HTTP_401_UNAUTHORIZED, status.HTTP_403_FORBIDDEN, status.HTTP_404_NOT_FOUND]
    ),
    dependencies=[Depends(requires_fab_custom_view("GET", permissions.RESOURCE_ROLE))],
)
def get_role(name: str = Path(..., min_length=1)) -> RoleResponse:
    """Get an existing role."""
    with get_application_builder():
        return FABAuthManagerRoles.get_role(name=name)


@roles_router.patch(
    "/roles/{name}",
    responses=create_openapi_http_exception_doc(
        [
            status.HTTP_400_BAD_REQUEST,
            status.HTTP_401_UNAUTHORIZED,
            status.HTTP_403_FORBIDDEN,
            status.HTTP_404_NOT_FOUND,
        ]
    ),
    dependencies=[Depends(requires_fab_custom_view("PATCH", permissions.RESOURCE_ROLE))],
)
def patch_role(
    body: RoleBody,
    name: str = Path(..., min_length=1),
    update_mask: str | None = Query(None, description="Comma-separated list of fields to update"),
) -> RoleResponse:
    """Update an existing role."""
    with get_application_builder():
        return FABAuthManagerRoles.patch_role(name=name, body=body, update_mask=update_mask)
